/*
 minimalistic driver that displays "Booting..." on an LCD connected to the
 gpio pins of a Soekris Net4501

 **********************************************************************
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 **********************************************************************
 *
 */
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>

/* Deal with CONFIG_MODVERSIONS */
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/proc_fs.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/fcntl.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/uaccess.h>


/* 8 gpio pins  */
#define NUMBER_OF_PINS 8

#define OUR_NAME                "writelcd"
#define VERSION                 "1.0.1"

/* defines for IO-Addresses/offsets */
#define SC520_PIOPFS15_0        0xC20
#define SC520_PIOPFS31_16       0xC22
#define SC520_CSPFS             0xC24
#define SC520_CLKSEL            0xC26
#define SC520_DSCTL             0xC28
#define SC520_PIODIR15_0        0xC2A
#define SC520_PIODIR15_8        0xC2B
#define SC520_PIODIR31_16       0xC2C
#define SC520_PIODIR31_24       0xC2D
#define SC520_PIODATA15_0       0xC30
#define SC520_PIODATA15_9       0xC31
#define SC520_PIODATA31_16      0xC32
#define SC520_PIODATA31_24      0xC33
#define SC520_PIOSET15_0        0xC34
#define SC520_PIOSET15_8        0xC35
#define SC520_PIOSET31_16       0xC36
#define SC520_PIOSET31_24       0xC37
#define SC520_PIOCLR15_0        0xC38
#define SC520_PIOCLR15_8        0xC39
#define SC520_PIOCLR31_16       0xC3A
#define SC520_PIOCLR31_24       0xC3B
#define SC520_CBAR              0xFFFC

#define SC520_MMIO_SIZE         0xC3C

#define MMCR_BASE_DEFAULT 0xFFFEF000
#define SIGNAL_ENABLE 0x40
#define SIGNAL_RS 0x10

#define UPDATE_INTERVAL (HZ/5+1)

//static struct timer_list timer;
static int ledstate;
static int leddirection;


int init_dummy_module(void);
void cleanup_dummy_module(void);


/* global variables */
static char *mmcr_virt_addr;


unsigned int read_gpio(int);
void write_gpio(int,unsigned int);


unsigned int read_gpio(int io_offset) {
        static unsigned long int data;
        static unsigned int return_value;

        /* fetch the current status  */
        data = readl(mmcr_virt_addr + io_offset );

        return_value=0;
        return_value |= ((data & 0x1000)     !=0 ? 0x80 : 0);  /* GPIO7->PIO12*/
        return_value |= ((data & 0x800)      !=0 ? 0x40 : 0);  /* GPIO6->PIO11*/
        return_value |= ((data & 0x400000)   !=0 ? 0x20 : 0);  /* GPIO5->PIO22*/
        return_value |= ((data & 0x200000)   !=0 ? 0x10 : 0);  /* GPIO4->PIO21*/
        return_value |= ((data & 0x100)      !=0 ? 0x08 : 0);  /* GPIO3->PIO8 */
        return_value |= ((data & 0x80)       !=0 ? 0x04 : 0);  /* GPIO2->PIO7 */
        return_value |= ((data & 0x40)       !=0 ? 0x02 : 0);  /* GPIO1->PIO6 */
        return_value |= ((data & 0x20)       !=0 ? 0x01 : 0);  /* GPIO0->PIO5 */

        return(return_value & 0xFF);

}

void write_gpio(int io_offset, unsigned int new_value) {
        static unsigned long int set_data;
        static unsigned long int clear_data;
        static unsigned long int data;

                if (io_offset == SC520_PIODIR15_0) {
                /* fetch the current status and get rid of the values we will set */
                /* this way we can simply "or" this mask with ne new values */
                /* mask: 1111 1111 1001 1111  1110 0110 0001 1111 */
                data = readl(mmcr_virt_addr + io_offset ) & 0xFF9FE61F;


                        // insert the new values
                        data |= ((new_value & 0x80)!=0)?0x1000:0;       // GPIO7 -> PIO12
                        data |= ((new_value & 0x40)!=0)?0x800 :0;       // GPIO6 -> PIO11

                        data |= ((new_value & 0x20)!=0)?0x400000 :0;    // GPIO5 -> PIO22
                        data |= ((new_value & 0x10)!=0)?0x200000 :0;    // GPIO4 -> PIO21

                        data |= ((new_value & 0x08)!=0)?0x100:0;        // GPIO3 -> PIO08
                        data |= ((new_value & 0x04)!=0)?0x80 :0;        // GPIO2 -> PIO07
                        data |= ((new_value & 0x02)!=0)?0x40 :0;        // GPIO1 -> PIO06
                        data |= ((new_value & 0x01)!=0)?0x20 :0;        // GPIO0 -> PIO05

                        //write the data
                        writel(data, mmcr_virt_addr + io_offset);
                } else {

                        // insert the new values
                        set_data=0;
                        clear_data=0;

                        if ((new_value & 0x80)!=0) set_data|=0x1000; else  clear_data|=0x1000; // GPIO7 -> PIO12
                        if ((new_value & 0x40)!=0) set_data|=0x800;  else  clear_data|=0x800; // GPIO6 -> PIO11

                        if ((new_value & 0x20)!=0) set_data|=0x400000; else  clear_data|=0x400000; // GPIO5 -> PIO22
                        if ((new_value & 0x10)!=0) set_data|=0x200000; else  clear_data|=0x200000; // GPIO4 -> PIO21

                        if ((new_value & 0x08)!=0) set_data|=0x100; else  clear_data|=0x100; // GPIO3 -> PIO08
                        if ((new_value & 0x04)!=0) set_data|=0x80;  else  clear_data|=0x80;  // GPIO2 -> PIO07
                        if ((new_value & 0x02)!=0) set_data|=0x40;  else  clear_data|=0x40;  // GPIO1 -> PIO06
                        if ((new_value & 0x01)!=0) set_data|=0x20;  else  clear_data|=0x20;  // GPIO0 -> PIO05


                        if (set_data!=0) writel(set_data, mmcr_virt_addr + SC520_PIOSET15_0 );
                        if (clear_data!=0) writel(clear_data, mmcr_virt_addr + SC520_PIOCLR15_0 );
                }

}



static void HD_nibble(unsigned char nibble)
{

        // clear ENABLE
        // put data on DB1..DB4
        // nibble already contains RS bit!
        write_gpio(SC520_PIODATA15_0, nibble);

        // Address set-up time
        udelay(1);

        // rise ENABLE
        write_gpio(SC520_PIODATA15_0, nibble | SIGNAL_ENABLE);

        // Enable pulse width
        udelay(1);

        // lower ENABLE
        write_gpio(SC520_PIODATA15_0, nibble);
}


static void HD_byte (unsigned char data, unsigned char RS)
{
        // send high nibble of the data
        HD_nibble (((data>>4)&0x0f)|RS);

        // Make sure we honour T_CYCLE
        udelay(1);

        // send low nibble of the data
        HD_nibble((data&0x0f)|RS);
}


static void HD_command (unsigned char cmd, int delay)
{
        HD_byte (cmd, 0);
        udelay(delay);
}


static void HD_write (char *string, int len, int delay)
{
    while (len--) {

                // send data with RS enabled
                HD_byte (*(string++), SIGNAL_RS);

                // wait for command completion
                udelay(delay);
        }
}

/*static void updateDisplay(unsigned long data)
{
        static char buf[9];
        int i;

        if (buf[0]==0) {
                // First loop;
                for (i=0;i<9;i++)
                {
                        buf[i]='-';
                }
        }

        buf[ledstate] = '-';
        ledstate++;
        if (ledstate>8) ledstate=0;
        buf[ledstate] = '>';

        // goto 7/0
        HD_command (0x80|11, 40);
        HD_write(buf,9,40);


        timer.expires = jiffies + UPDATE_INTERVAL;
        add_timer(&timer);
}*/



/* startup/cleanup ---------------------------------------------------------*/
/* Initialize the module - register the proc file */
int init_dummy_module()
{
        static unsigned long cbar;
        static unsigned long mmcrbase;

        /* find MMCR base the address */
        cbar = inl_p(SC520_CBAR);
        /*printk(OUR_NAME ": CBAR: 0x%08lx\n", cbar);*/
        /* check if MMCR aliasing bit is set */
        if (cbar & 0x80000000) {
                /*printk(OUR_NAME ": MMCR Aliasing enabled.\n");*/
                mmcrbase = cbar & 0x3fffffff;
        } else {
                printk(OUR_NAME "!!! WARNING !!!\n"
                  "\t MMCR Aliasing found NOT enabled!\n"
                  "\t Using default value of: %x\n"
                  , MMCR_BASE_DEFAULT
                  );
                mmcrbase = MMCR_BASE_DEFAULT;
        }

        /* map bus memory into CPU space   */
        mmcr_virt_addr = ioremap(mmcrbase, SC520_MMIO_SIZE);

        if (!mmcr_virt_addr) {
                printk(OUR_NAME ": Could not remap memory. Terminating\n");
                return -ENOMEM;

        }

        /* set GPIO0-7 to output*/
        write_gpio(SC520_PIODIR15_0, 0xFF);

                // Init the LCD
                HD_nibble(0x03); udelay(100);mdelay(4); // 4 Bit mode, wait 4.1 ms
                HD_nibble(0x03); udelay(100);  // 4 Bit mode, wait 100 us
                HD_nibble(0x03); udelay(100);mdelay(4); // 4 Bit mode, wait 4.1 ms
                HD_nibble(0x02); udelay(100);  // 4 Bit mode, wait 100 us
                HD_command (0x28, 40);         // 4 Bit mode, 1/16 duty cycle, 5x8 font


                HD_command (0x08, 40);   // Display off, cursor off, blink off
                //HD_command (0x0c, 640); mdelay(1);        // Display on, cursor off, blink off, wait 1.64 ms
                HD_command (0x0F, 640); mdelay(1);        // Display on, cursor on, blink on, wait 1.64 ms
                HD_command (0x06, 40);   // cursor moves to right, no shift

                  HD_command (0x01, 640);mdelay(1); // clear display

                mdelay(5);

                //goto 0/0
                HD_command ((0x80), 40);

                // Now send our string
                HD_write("Booting...",10,40);
                ledstate = 7;
                leddirection=1;
                //init_timer(&timer);
                //timer.function = updateDisplay;
                //timer.data = 0;
                //timer.expires = 0; /* pacify test of jiffies < timer.expired */
                //mod_timer(&timer, jiffies + UPDATE_INTERVAL);

                printk(OUR_NAME ": Version " VERSION " (C) 2003 Martin Hejl\n");
        return 0;
}


/* Cleanup - unregister our file from /proc */
void cleanup_dummy_module()
{
                //del_timer_sync(&timer);

                HD_command (0x01, 640);mdelay(1); // clear display
        iounmap(mmcr_virt_addr);
}

module_init(init_dummy_module);
module_exit(cleanup_dummy_module);

MODULE_AUTHOR("Martin Hejl");
MODULE_DESCRIPTION("Soekris net4501 GPIO LCD driver");
MODULE_LICENSE("GPL");
EXPORT_NO_SYMBOLS;
