/*
 Soekris Net4801 GPIO driver
 Provides basic control (read and write) over the 12 general purpose
 IO-pins of the Soekris Net4801 and the error-led (which happens to
 be connected to another general purpose IO pin)

 **********************************************************************
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 **********************************************************************
 *
 */


#include "4801driver.h"
static int net4801_init(void);
static void net4801_cleanup(void);


extern struct file_operations gpio_dispatch_fops;
extern int procfile_gpio_read(char*,char**,off_t,int,int*,void*);
extern int procfile_gpio_write(struct file*,const char*,unsigned long,void*);
extern int procfile_led_read(char*,char**,off_t,int,int*,void*);
extern int procfile_led_write(struct file*,const char*,unsigned long,void*);
extern int procfile_settings_read(char*,char**,off_t,int,int*,void*);
extern int procfile_settings_write(struct file*,const char*,unsigned long,void*);

static void toString(unsigned long value, char* buffer, int number_of_bits) {
        static int i;

        /* convert it into a string */
        for(i=number_of_bits;i>0;i--){
                buffer[number_of_bits-i]=test_bit(i-1,&value)?'1':'0';
        }

        buffer[number_of_bits] = '\n';
        buffer[number_of_bits+1] = 0;

}

static unsigned long  fromString(char* buffer, int number_of_bits) {
        static int i;
        static unsigned long ret_val=0;

        ret_val = 0;
        /* Create WORD to write from the string */
        for(i=0;i<number_of_bits;i++){
                if (buffer[i] == 0) break;

                if (buffer[i] == '0' || buffer[i] == '1') {
                        ret_val=ret_val<<1;
                }

                if (buffer[i] == '1') {
                        ret_val |= 1;

                }
        }

        return(ret_val);


}



static void _net4801_write_error_led(unsigned int on) {

        unsigned char value=0;
        value = inb(geode_gpio_base+GEODE_GPIO_DATA_OUT+2);
        if (on!=0) {
                set_bit(4, &value);
        } else {
                clear_bit(4, &value);
        };
        outb(value,geode_gpio_base+GEODE_GPIO_DATA_OUT+2);
}

static unsigned int _net4801_read_error_led() {
        return ((inl(geode_gpio_base+GEODE_GPIO_DATA_IN)&0x80000)!=0?1:0);
}

static void _net4801_select_pin(unsigned short port, unsigned short pin) {
                /* select port/pin */
                outb(SIO_GPIO_PIN_SELECT,SIO_INDEX);
                outb(((port & 0xF)<<4) | (pin & 0xF), SIO_DATA);
}

static void _net4801_write_config(unsigned short port, unsigned short pin, unsigned char setting) {

        _net4801_select_pin(port, pin);

                /* configure selected port/pin */
                outb(SIO_GPIO_PIN_CONFIGURE,SIO_INDEX);

        outb(setting, SIO_DATA);
}

static unsigned char _net4801_read_config(unsigned short port, unsigned short pin) {

        _net4801_select_pin(port, pin);

                /* configure selected port/pin */
                outb(SIO_GPIO_PIN_CONFIGURE,SIO_INDEX);

        return (inb(SIO_DATA));
}

static void _net4801_writeGPIO07(unsigned char new_value) {
        outb(new_value, SIO_GPIO_BANK2);
}

static unsigned char _net4801_readGPIO07() {
        return inb(SIO_GPIO_BANK2);
}


static void _net4801_setGPIO07Direction(unsigned char new_direction) {
        int pin;

        for (pin=0;pin<=7;pin++) {
                                if (((new_direction>>pin)&1) != 0)
                        _net4801_write_config(2,pin,CONFIG_GPIO_OUTPUT);
                else
                        _net4801_write_config(2,pin,CONFIG_GPIO_INPUT);
                }
}


static unsigned char _net4801_getGPIO07Direction() {
        int pin;
        unsigned char value=0;

        for (pin=0;pin<=7;pin++) {
                        value |= ((char)(_net4801_read_config(2,pin)) & 0x01)<<pin;
                }
                return (value);
}

static void _net4801_writeGPIO(unsigned int new_value) {
        /* we can safely ignore the direction since the spec explicitly
        states that writing to a GPIO set to input will do nothing */

        outb((char)((new_value & 0x300)>>4), SIO_GPIO_BANK0);
        outb((char)((new_value & 0x400)>>7 | ((new_value & 0x800)>>9)), SIO_GPIO_BANK1);
        outb((char)(new_value & 0xFF), SIO_GPIO_BANK2);
}

static unsigned int _net4801_readGPIO() {
        unsigned char value=0;
        unsigned long ret_val =0;

        ret_val = inb(SIO_GPIO_BANK2);
        value = inb(SIO_GPIO_BANK1);
        ret_val |= (value & 0x8)<<7;
        ret_val |= (value & 0x4)<<9;

        value = inb(SIO_GPIO_BANK0);
        ret_val |= (value & 0x30)<<4;

        return (ret_val);
}




static void _net4801_setGPIODirection(unsigned int new_direction) {
        int pin;

        for (pin=0;pin<=7;pin++) {
                        _net4801_write_config(2,pin,test_bit(pin, &new_direction)?CONFIG_GPIO_OUTPUT:CONFIG_GPIO_INPUT);
                }

                _net4801_write_config(0,4,(new_direction&0x100) != 0?CONFIG_GPIO_OUTPUT:CONFIG_GPIO_INPUT);
                _net4801_write_config(0,5,(new_direction&0x200) != 0?CONFIG_GPIO_OUTPUT:CONFIG_GPIO_INPUT);

                _net4801_write_config(1,3,(new_direction&0x400) != 0?CONFIG_GPIO_OUTPUT:CONFIG_GPIO_INPUT);
                _net4801_write_config(1,2,(new_direction&0x800) != 0?CONFIG_GPIO_OUTPUT:CONFIG_GPIO_INPUT);
}


static unsigned int _net4801_getGPIODirection() {
        int pin;
        unsigned int value=0;
        unsigned int config;

        for (pin=0;pin<=7;pin++) {
                        config = _net4801_read_config(2,pin);
                        if ((config & 0x01)!=0) set_bit(pin, &value);
                }


                if ((_net4801_read_config(0,4) & 0x01)!=0) set_bit(8,&value);
                if ((_net4801_read_config(0,5) & 0x01)!=0) set_bit(9,&value);
                if ((_net4801_read_config(1,3) & 0x01)!=0) set_bit(10,&value);
                if ((_net4801_read_config(1,2) & 0x01)!=0) set_bit(11,&value);

                return (value&0xFFF);
}


static struct gpio_operations driver_ops =  {
        .writeErrorLed = _net4801_write_error_led,
        .readErrorLed        = _net4801_read_error_led,
        .write8Bit                = _net4801_writeGPIO07,
        .read8Bit                = _net4801_readGPIO07,
        .set8BitDirection        = _net4801_setGPIO07Direction,
        .get8BitDirection        = _net4801_getGPIO07Direction,
        .write16Bit        = _net4801_writeGPIO,
        .read16Bit                 = _net4801_readGPIO,
        .set16BitDirection = _net4801_setGPIODirection,
        .get16BitDirection = _net4801_getGPIODirection,
        .toProcString        = toString,
        .fromProcString = fromString,
        .init                        = net4801_init,
        .cleanup                = net4801_cleanup
};

static int net4801_init(void)
{
        struct pci_dev *bridge;
        unsigned base;

        printk(OUR_NAME ": Version " VERSION " (C) 2003 Martin Hejl\n");

        if ((bridge = pci_find_device(SC1100_VENDOR_ID,
                                  SC1100_F0_DEVICE_ID ,
                                  NULL)) == NULL)
                return -ENODEV;

        base = pci_resource_start(bridge, 0);
        printk(KERN_INFO OUR_NAME ": Geode GPIO base 0x%x\n", base);

        if (request_region(base, GEODE_GPIO_SIZE, "Soekris net4801 GPIO") == 0) {
                printk(KERN_ERR OUR_NAME ": can't allocate I/O for Geode GPIOs\n");
                return -EBUSY;
        }

        geode_gpio_base = base;


        offset_87336 = 0;
        outb(SIO_SID,SIO_INDEX);
        if (inb(SIO_DATA) != SIO_SID_VALUE)
        {
                offset_87336 = SIO_BASE_OFFSET;
                outb(SIO_SID,SIO_INDEX);
                if (inb(SIO_DATA) != SIO_SID_VALUE)
                {
                        printk(KERN_ERR OUR_NAME ": PC87336 not detected\n" );

                        /* clean up */
                        unregister_chrdev(gpio_major, OUR_NAME);
                        release_region(geode_gpio_base, GEODE_GPIO_SIZE);
                        return -EBUSY;
                }
        }


        outb_p(SIO_DEV_SEL, SIO_INDEX);
        outb_p(SIO_GPIO_DEV, SIO_DATA);        /* Talk to GPIO regs. */

        outb_p(SIO_DEV_ENB, SIO_INDEX);
        if (inb_p(SIO_DATA)==0) {
                printk(KERN_ERR OUR_NAME ": GPIOs are disabled\n");
                /* clean up */
                unregister_chrdev(gpio_major, OUR_NAME);
                release_region(geode_gpio_base, GEODE_GPIO_SIZE);
                return -ENODEV;
        }

        // read the base address
        outb(SIO_BASE_HADDR,SIO_INDEX);
        sio_gpio_base = inb(SIO_DATA) << 8;

        outb(SIO_BASE_LADDR,SIO_INDEX);
        sio_gpio_base |= inb(SIO_DATA) ;

        return 0;
}

static void net4801_cleanup(void)
{
        release_region(geode_gpio_base, GEODE_GPIO_SIZE);

}

